@@ -1,15 +1,16 @@
 """EQL syntax tree nodes/schema."""
 from __future__ import unicode_literals
 
-import datetime
 import re
 from collections import OrderedDict
-from operator import lt, le, eq, ne, ge, gt, mul, truediv, mod, add, sub
+from operator import lt, le, eq, ne, ge, gt
 from string import Template
+from enum import Enum
 
+from .errors import EqlError
 from .signatures import SignatureMixin
-from .types import STRING, BOOLEAN, NUMBER, NULL, PRIMITIVES
-from .utils import to_unicode, is_string, is_number, ParserConfig
+from .types import TypeHint, NodeInfo  # noqa: F401
+from .utils import to_unicode, is_string, is_number, ParserConfig, fold_case
 
 __all__ = (
     # base classes
@@ -24,6 +25,7 @@ __all__ = (
     "Null",
     "Boolean",
     "TimeRange",
+    "TimeUnit",
 
     # fields and subfields
     "Field",
@@ -31,6 +33,8 @@ __all__ = (
     # boolean logic
     "Comparison",
     "InSet",
+    "IsNotNull",
+    "IsNull",
     "And",
     "Or",
     "Not",
@@ -77,6 +81,20 @@ class BaseNode(object):
         for key in self.__slots__:
             yield key, getattr(self, key, None)
 
+    def children(self):  # type: () -> list[BaseNode]
+        """Collect over all child nodes."""
+        children = []
+
+        def recurse(descendant):
+            if isinstance(descendant, BaseNode):
+                children.append(descendant)
+            elif hasattr(descendant, "__iter__"):
+                for c in descendant:
+                    recurse(c)
+
+        recurse(c for c in self.iter_slots())
+        return children
+
     def optimize(self, recursive=False):
         """Optimize an AST."""
         return Optimizer(recursive=recursive).walk(self)
@@ -157,26 +175,59 @@ class Expression(EqlNode):
 
     precedence = 0
 
+    def foldable(self):
+        """Determine if an expression can be folded."""
+        return all(c.foldable() for c in self.children())
+
+    def fold(self):
+        """Fold an expression."""
+        optimized = self.optimize(recursive=True)
+        if isinstance(optimized, Literal):
+            return optimized.value
+
+        raise EqlError("Unable to fold expression: {}".format(self))
+
     def __and__(self, other):
         """Boolean AND between two AST nodes."""
-        if isinstance(other, Literal):
-            if other.value:
-                return self
+        if self == Boolean(False) or other == Boolean(False):
             return Boolean(False)
+        elif self == Boolean(True):
+            return other
+        elif other == Boolean(True):
+            return self
 
-        if isinstance(other, And):
-            return And([self] + other.terms)
-        return And([self, other])
+        if self == Null() or other == Null():
+            return Null()
+
+        # flatten out the terms in the And
+        terms = []
+        for t in (self, other):
+            if isinstance(t, And):
+                terms.extend(t.terms)
+            else:
+                terms.append(t)
+        return And(terms)
 
     def __or__(self, other):
         """"Boolean OR between two AST nodes."""
-        if isinstance(other, Literal):
-            if other.value:
-                return Boolean(True)
+        if self == Boolean(True) or other == Boolean(True):
+            return Boolean(True)
+        elif self == Boolean(False):
+            return other
+        elif other == Boolean(False):
+            return self
 
-        if isinstance(other, Or):
-            return Or([self] + other.terms)
-        return Or([self, other])
+        if self == Null() or other == Null():
+            return Null()
+
+        # flatten out the terms in the Or
+        terms = []
+        for t in (self, other):
+            if isinstance(t, Or):
+                terms.extend(t.terms)
+            else:
+                terms.append(t)
+        return Or(terms)
 
     def __invert__(self):
         """Negate an expression with Not."""
@@ -188,7 +239,7 @@ class Literal(Expression):
 
     __slots__ = 'value',
     precedence = Expression.precedence + 1
-    type_hint = PRIMITIVES
+    type_hint = None  # type: TypeHint
 
     def __init__(self, value):
         """Create an EQL value from a python value."""
@@ -208,7 +259,7 @@ class Literal(Expression):
         elif is_string(python_value):
             return String
         else:
-            raise TypeError("Unable to convert python value to a literal.")
+            raise TypeError("Unable to convert {} to a literal.".format(python_value))
 
     @classmethod
     def from_python(cls, python_value):
@@ -216,24 +267,6 @@ class Literal(Expression):
         subcls = cls.find_type(python_value)
         return subcls(python_value)
 
-    def __and__(self, other):
-        """Shortcut ANDing of Static Value nodes together."""
-        if isinstance(other, Literal):
-            return Boolean(self.value and other.value)
-        elif self.value:
-            return other
-        else:
-            return Boolean(False)
-
-    def __or__(self, other):
-        """Shortcut ORing of Static Value nodes together."""
-        if isinstance(other, Literal):
-            return Boolean(self.value or other.value)
-        elif self.value:
-            return self
-        else:
-            return other
-
     def __invert__(self):
         """Negate a static value."""
         return Boolean(not self.value)
@@ -242,7 +275,7 @@ class Literal(Expression):
 class Boolean(Literal):
     """Boolean literal."""
 
-    type_hint = BOOLEAN
+    type_hint = TypeHint.Boolean
 
     def _render(self):
         return 'true' if self.value else 'false'
@@ -251,12 +284,16 @@ class Boolean(Literal):
 class Null(Literal):
     """Null literal."""
 
-    type_hint = NULL
+    type_hint = TypeHint.Null
 
     def __init__(self, value=None):
         """Null literal value."""
         super(Null, self).__init__(None)
 
+    def __invert__(self):
+        """Null values can't be inverted."""
+        return Null()
+
     def _render(self):
         return 'null'
 
@@ -264,7 +301,7 @@ class Null(Literal):
 class Number(Literal):
     """Numeric literal."""
 
-    type_hint = NUMBER
+    type_hint = TypeHint.Numeric
 
     def _render(self):
         return to_unicode(self.value)
@@ -285,7 +322,7 @@ class String(Literal):
     }
     reverse_patterns = {v: k for k, v in escape_patterns.items()}
     escape_re = r'[{}]'.format('|'.join(escape_patterns.values()))
-    type_hint = STRING
+    type_hint = TypeHint.String
 
     @classmethod
     def escape(cls, s):
@@ -305,48 +342,57 @@ class String(Literal):
         return '"{}"'.format(self.escape(self.value))
 
 
+class TimeUnit(Enum):
+    """Elasticsearch compatible time units."""
+
+    Milliseconds = "ms"
+    Seconds = "s"
+    Minutes = "m"
+    Hours = "h"
+    Days = "d"
+
+    @classmethod
+    def get_units(cls):
+        """Return all time units in order of precision."""
+        return [cls.Milliseconds, cls.Seconds, cls.Minutes, cls.Hours, cls.Days]
+
+    def as_milliseconds(self):
+        """Convert a time unit to a number of milliseconds."""
+        if self == TimeUnit.Milliseconds:
+            return 1
+        elif self == TimeUnit.Seconds:
+            return 1000
+        elif self == TimeUnit.Minutes:
+            return 1000 * 60
+        elif self == TimeUnit.Hours:
+            return 1000 * 60 * 60
+        elif self == TimeUnit.Days:
+            return 1000 * 60 * 60 * 24
+        else:
+            raise ValueError("Unknown time unit {}".format(repr(self)))
+
+    def __repr__(self):
+        """Override the default __repr__ method."""
+        return "{}({})".format(type(self).__name__, repr(self.value))
+
+
 class TimeRange(Expression):
     """EQL node for an interval of time."""
 
-    __slots__ = 'delta',
+    __slots__ = 'quantity', 'unit',
     precedence = Expression.precedence + 1
 
-    def __init__(self, delta):  # type: (datetime.timedelta) -> None
+    def __init__(self, quantity, unit):  # type: (int, TimeUnit) -> None
         """EQL time interval."""
-        self.delta = delta
+        self.quantity = quantity
+        self.unit = unit
 
-    @classmethod
-    def convert(cls, node):
-        """Convert a StaticValue to a time range."""
-        if isinstance(node, TimeRange):
-            return node
-        elif isinstance(node, Number):
-            return TimeRange(datetime.timedelta(seconds=node.value))
+    def as_milliseconds(self):
+        """Convert a time range to milliseconds."""
+        return self.quantity * self.unit.as_milliseconds()
 
     def _render(self):
-        interval = self.delta.total_seconds()
-        second = 1
-        minute = 60 * second
-        hour = minute * 60
-        day = hour * 24
-        decimal = interval
-        unit = 's'
-
-        if interval >= day:
-            decimal = float(interval) / day
-            unit = 'd'
-        elif interval >= hour:
-            decimal = float(interval) / hour
-            unit = 'h'
-        elif interval >= minute:
-            if interval % minute == 0 or (interval % second) != 0:
-                decimal = float(interval) / minute
-                unit = 'm'
-
-        # Drop fractional part if it's 0
-        if decimal == int(decimal):
-            decimal = int(decimal)
-        return '{}{}'.format(decimal, unit)
+        return '{:d}{:s}'.format(self.quantity, self.unit.value)
 
 
 class Field(Expression):
@@ -357,6 +403,8 @@ class Field(Expression):
     __slots__ = 'base', 'path',
     precedence = Expression.precedence + 1
 
+    field_re = re.compile("^[_A-Za-z][_A-Za-z0-9]*$")
+
     def __init__(self, base, path=None):
         """Query the event for the field expression.
 
@@ -378,14 +426,24 @@ class Field(Expression):
         """Get the full path for a field."""
         return [self.base] + self.path
 
+    @classmethod
+    def escape_ident(cls, key):
+        """Escape identifiers that are keywords."""
+        from .parser import keywords
+
+        if key in keywords or cls.field_re.match(key) is None:
+            return "`{key}`".format(key=key)
+        return key
+
     def _render(self):
-        text = self.base
+        text = [self.escape_ident(self.base)]
+
         for key in self.path:
             if is_number(key):
-                text += "[{}]".format(key)
+                text.append("[{}]".format(key))
             else:
-                text += ".{}".format(key)
-        return text
+                text.append(".{}".format(self.escape_ident(key)))
+        return "".join(text)
 
 
 class FunctionCall(Expression):
@@ -470,7 +528,6 @@ class MathOperation(Expression):
     __slots__ = 'left', 'operator', 'right'
     OPERATORS = ('*', '/', '%', '+', '-')
 
-    op_lookup = {'*': mul, '/': truediv, '%': mod, '+': add, '-': sub}
     func_lookup = {"*": "multiply", "+": "add", "-": "subtract", "%": "modulo", "/": "divide"}
 
     min_precedence = NamedSubquery.precedence + 1
@@ -501,11 +558,6 @@ class MathOperation(Expression):
         """Make the template dynamic."""
         return self.negative_template if self.left == Number(0) else self.full_template
 
-    @property
-    def func(self):
-        """Get a callback function for the specific operator."""
-        return self.op_lookup[self.operator]
-
 
 class Comparison(Expression):
     """Represents a comparison between two values, as in ``<expr> <comparator> <expr>``.
@@ -555,6 +607,30 @@ class Comparison(Expression):
         return super(Comparison, self).__and__(other)
 
 
+class IsNull(Expression):
+    """Node for checking if values are null."""
+
+    template = Template("$expr == null")
+    precedence = Comparison.precedence
+    __slots__ = "expr",
+
+    def __init__(self, expr):  # type: (Expression) -> None
+        """Check if a value is null."""
+        self.expr = expr
+
+
+class IsNotNull(Expression):
+    """Node for checking if values are null."""
+
+    template = Template("$expr != null")
+    precedence = Comparison.precedence
+    __slots__ = "expr",
+
+    def __init__(self, expr):  # type: (Expression) -> None
+        """Check if a value is not null."""
+        self.expr = expr
+
+
 class InSet(Expression):
     """Check if the value of a field within an event matches a list of values."""
 
@@ -584,7 +660,7 @@ class InSet(Expression):
                 continue
             k = literal.value
             if isinstance(literal, String):
-                values.setdefault(k.lower(), literal)
+                values.setdefault(fold_case(k), literal)
             else:
                 values[k] = literal
         return values
@@ -712,11 +788,11 @@ class Not(Expression):
 
     def demorgans(self):
         """Apply DeMorgan's law."""
-        if isinstance(self, Or):
-            return And([~ t for t in self.terms]).optimize()
+        if isinstance(self.term, Or):
+            return And([(~ t).optimize() for t in self.term.terms]).optimize()
 
-        elif isinstance(self, And):
-            return Or([~ t for t in self.terms]).optimize()
+        elif isinstance(self.term, And):
+            return Or([(~ t).optimize() for t in self.term.terms]).optimize()
 
         else:
             return ~ self.term.optimize()
@@ -743,15 +819,6 @@ class And(BaseCompound):
     precedence = Not.precedence + 1
     operator = 'and'
 
-    def __and__(self, other):
-        """Flatten multiple ``and`` terms."""
-        terms = self.terms
-        if isinstance(other, And):
-            terms.extend(other.terms)
-        else:
-            terms.append(other)
-        return And(terms)
-
 
 class Or(BaseCompound):
     """Perform a boolean ``or`` on a list of expressions."""
@@ -759,15 +826,6 @@ class Or(BaseCompound):
     precedence = And.precedence + 1
     operator = 'or'
 
-    def __or__(self, other):
-        """Flatten multiple ``or`` terms."""
-        terms = self.terms
-        if isinstance(other, Or):
-            terms.extend(other.terms)
-        else:
-            terms.append(other)
-        return Or(terms)
-
 
 class EventQuery(EqlNode):
     """Query over a specific event type with a boolean condition."""
@@ -811,18 +869,26 @@ class NamedParams(EqlNode):
 class SubqueryBy(EqlNode):
     """Node for holding the :class:`~EventQuery` and parameters to join on."""
 
-    __slots__ = 'query', 'params', 'join_values'
+    __slots__ = 'query', 'join_values', 'fork',
 
-    def __init__(self, query, params=None, join_values=None):
+    def __init__(self, query, join_values=None, fork=None):
         """Init.
 
         :param EventQuery query: The event query enclosed in the term
-        :param NamedParams params: The parameters for the query.
         :param list[Expression] join_values: The field to join values on
+        :param bool fork: Toggle for copying instead of moving a sequence on match
         """
         self.query = query
-        self.params = params or NamedParams()
         self.join_values = join_values or []
+        self.fork = fork
+
+    @property
+    def params(self):
+        """Keep params for backwards compatibility."""
+        params = {}
+        if self.fork is not None:
+            params["fork"] = Boolean(self.fork)
+        return NamedParams(params)
 
     def _render(self):
         text = "[{}]".format(self.query.render())
@@ -864,24 +930,31 @@ class Sequence(EqlNode):
     Sequence supports the ``until`` keyword, which indicates an event that causes it to terminate early.
     """
 
-    __slots__ = 'queries', 'params', 'close'
+    __slots__ = 'queries', 'max_span', 'close'
 
-    def __init__(self, queries, params=None, close=None):
+    def __init__(self, queries, max_span=None, close=None):
         """Create a Sequence of multiple events.
 
         :param list[SubqueryBy] queries: List of queries to be sequenced
-        :param NamedParams params: Dictionary of timing parameters for the sequence.
+        :param TimeUnit max_span: Dictionary of timing parameters for the sequence.
         :param SubqueryBy close: An optional query that causes all sequence state to expire
         """
         self.queries = queries
-        self.params = params or NamedParams()
+        self.max_span = max_span
         self.close = close
 
+    @property
+    def params(self):
+        """Keep params for backwards compatibility."""
+        params = {}
+        if self.max_span is not None:
+            params["maxspan"] = self.max_span
+        return NamedParams(params)
+
     def _render(self):
         text = 'sequence'
-        params = self.params.render()
-        if params:
-            text += ' with {}'.format(self.params.render())
+        if self.max_span:
+            text += ' with maxspan={}'.format(self.max_span.render())
         text += '\n'
         text += self.indent('\n'.join(query.render() for query in self.queries))
 
@@ -915,8 +988,8 @@ class PipeCommand(EqlNode, SignatureMixin):
         return decorator
 
     @classmethod
-    def output_schemas(cls, arguments, type_hints, event_schemas):
-        # type: (list, list, list[Schema]) -> list[Schema]
+    def output_schemas(cls, arguments, event_schemas):
+        # type: (list[NodeInfo], list[Schema]) -> list[Schema]
         """Output a list of schemas for each event in the pipe."""
         return event_schemas
 
@@ -1107,7 +1180,7 @@ class PreProcessor(ParserConfig):
 
             def _walk_field(self, node, *args, **kwargs):
                 if node.base in self.preprocessor.constants and not node.path:
-                        return self.preprocessor.constants[node.base].value
+                    return self.preprocessor.constants[node.base].value
                 return self._walk_base_node(node, *args, **kwargs)
 
             def _walk_function_call(self, node, *args, **kwargs):
